#include "basics.h"
#include "Py.h"
#include "Plugin.h"

#include "ALSA.h"
#include "ALSA_Seq.h"

namespace ALSA {

SeqClient::SeqClient()
{
	handle = 0;
	q = -1;
	port = -1;
	client = -1;
	direction = -1;
}

SeqClient::~SeqClient()
{
	if (handle)
	{
		if (port != -1)
			snd_seq_delete_simple_port (handle, port);
		
		if (q != -1)
			snd_seq_free_queue (handle, q);
		
		snd_seq_close (handle);
	}
}

void
SeqClient::open (char * name, int _direction)
{
	direction = _direction;

	ALSA_CALL (snd_seq_open (&handle, "hw", direction, SND_SEQ_NONBLOCK));
	
	q = snd_seq_alloc_queue (handle);
	client = snd_seq_client_id (handle);

	char s[64];
	PyObject * o = this->as_py_object();
	snprintf (s, 64, "<%s at %p>", o->ob_type->tp_name, o);
	s[63] = 0;
	snd_seq_set_client_name (handle, s);

	int port_caps = 0;

	if (direction == SND_SEQ_OPEN_OUTPUT)
		port_caps |= SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ;
	else
		port_caps |= SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE;

	port = snd_seq_create_simple_port (handle, name, port_caps,
			SND_SEQ_PORT_TYPE_MIDI_GENERIC);
	
	if (port < 0)
		throw IOError (-port);
}

void
SeqClient::subscribe (int major, int minor, bool connect)
{
	snd_seq_port_subscribe_t * subscription;
	snd_seq_port_subscribe_alloca (&subscription); /* fixme: leak */
	
	snd_seq_port_subscribe_set_queue (subscription, q);

	snd_seq_addr_t 
		me = {client, port},
		he = {major, minor};
		
	if (direction == SND_SEQ_OPEN_OUTPUT)
	{
		snd_seq_port_subscribe_set_sender (subscription, &me);
		snd_seq_port_subscribe_set_dest (subscription, &he);
	}
	else
	{
		snd_seq_port_subscribe_set_sender (subscription, &he);
		snd_seq_port_subscribe_set_dest (subscription, &me);
	}

	if (connect)
	{
		ALSA_CALL (snd_seq_subscribe_port ((snd_seq_t *) handle, subscription));
	}
	else
	{
		ALSA_CALL (snd_seq_unsubscribe_port ((snd_seq_t *) handle, subscription));
	}
}

PyObject *
SeqClient::_subscribe (PyObject * args)
{
	int maj, min;

	if (!PyArg_ParseTuple (args, "ii", &maj, &min))
		return 0;

	subscribe (maj, min, true);
	return PyNone();
}

PyObject *
SeqClient::_unsubscribe (PyObject * args)
{
	int maj, min;

	if (!PyArg_ParseTuple (args, "ii", &maj, &min))
		return 0;

	subscribe (maj, min, true);
	return PyNone();
}


/* //////////////////////////////////////////////////// */

SeqIn::SeqIn()
{ }

void
SeqIn::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"name", 0};
	char * name;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &name))
		throw blewit();

	open (name, SND_SEQ_OPEN_INPUT);

	int n = snd_seq_poll_descriptors_count (handle, POLLIN);
	if (n != 1)
		throw IOError() + "can only handle a single file descriptor";
		
	struct pollfd pfd;
	int d = snd_seq_poll_descriptors (handle, &pfd, 1, POLLIN);
	if (n != d)
		throw IOError() + "want 1 file descriptor, got ", d;

	if (plugin.fd >= 0 && plugin.fd != pfd.fd)
		throw AssertionError() + "cannot change the file descriptor";

	plugin.fd = pfd.fd;

	new_tx<TX> ("MIDI");
}

void
SeqIn::handle_io (int fd, IOCycle & c)
{
	/* 1 = fetch events from kernel. */
	int n = snd_seq_event_input_pending (handle, 1);
	if (n <= 0)
	{
		msg(), this, ": no events pending";
		return;
	}

	while (n--)
	{
		snd_seq_event_t * packet;
		ALSA_CALL (snd_seq_event_input ((snd_seq_t *) handle, &packet));
		
		if (packet->type == SND_SEQ_EVENT_SYSEX)
		{
			Packet * p = c.packet ((Event::Type) 0xF0, c.tick);
			byte * data = (byte *) packet + sizeof (snd_seq_event_t);
			p->set_data (data, packet->data.ext.len);
			c.push (p, tx(0));
		}
		else
		{
			Event e;
			e.type = Event::Invalid;
			e.t = c.tick;

			switch (packet->type)
			{
				case SND_SEQ_EVENT_NOTEOFF:
					e.set_type (0x80 | packet->data.note.channel);
					e.data[0] = packet->data.note.note;
					e.data[1] = packet->data.note.velocity;
					break;

				case SND_SEQ_EVENT_NOTEON:
					e.set_type (0x90 | packet->data.note.channel);
					e.data[0] = packet->data.note.note;
					e.data[1] = packet->data.note.velocity;
					break;

				case SND_SEQ_EVENT_KEYPRESS:
					e.set_type (0xA0 | packet->data.note.channel);
					e.data[0] = packet->data.note.note;
					e.data[1] = packet->data.note.velocity;
					break;

				case SND_SEQ_EVENT_CONTROLLER:
					e.set_type (0xB0 | packet->data.control.channel);
					e.data[0] = packet->data.control.param;
					e.data[1] = packet->data.control.value;
					break;

				case SND_SEQ_EVENT_PGMCHANGE:
					e.set_type (0xC0 | packet->data.control.channel);
					e.data[0] = packet->data.control.value;
					break;

				case SND_SEQ_EVENT_CHANPRESS:
					e.set_type (0xD0 | packet->data.control.channel);
					e.data[0] = packet->data.control.value;
					break;

				case SND_SEQ_EVENT_PITCHBEND:
					e.set_type (0xE0 | packet->data.control.channel);
					{
						int p = packet->data.control.value + 0x2000;
						e.data[1] = (p >> 7) & 0x7F;
						e.data[0] = p & 0x7F;
					}
					break;

				default:
					break;
			}

			if (e.type != Event::Invalid)
				c.push (e, tx(0));
		}

		/* clean up. */
		snd_seq_free_event (packet);
	}
}

/* ////////////////////////////////////////////////////////// */

SeqOut::SeqOut()
{ }

void
SeqOut::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"name", 0};
	char * name;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &name))
		throw blewit();

	open (name, SND_SEQ_OPEN_OUTPUT);

	snd_seq_ev_clear (&packet);
	snd_seq_ev_set_source (&packet, port);

	packet.queue = q;

	packet.type = 0;
	packet.flags = SND_SEQ_TIME_STAMP_TICK | SND_SEQ_TIME_MODE_REL;
	packet.time.tick = 0;

	snd_seq_ev_set_subs (&packet);

	plugin.flags = Cycle::Time;
	for (int i = 0; i < 16; ++i)
	{
		char s[8];
		sprintf (s, "%d", i + 1);
		new_rx<RX> (s);
	}
}

void
SeqOut::cycle (Cycle & c)
{
	int r;
	Packet * p = next_packet (c.tick.now, &r);

	while (p)
	{
		if (p->type == 0xF0)
			send_sysex (p);
		else
			send_packet (p, r);
		c.discard (p);
		p = next_packet (c.tick.now, &r);
	}
}

void
SeqOut::send_packet (Packet * p, int channel)
{
	snd_seq_ev_set_fixed (&packet);

	int status;
	
	if (p->type < 0x10)
		status = p->type << 4;
	else
	{
		channel = p->type & 0xF;
		status = p->type & Event::Mask & ~0xF;
	}

	switch (status)
	{
		case 0x80:
			packet.type = SND_SEQ_EVENT_NOTEOFF;
			packet.data.note.channel = channel;
			packet.data.note.note = p->data[0];
			packet.data.note.velocity = p->data[1];
			packet.data.note.off_velocity = p->data[1];
			break;

		case 0x90:
			packet.type = SND_SEQ_EVENT_NOTEON;
			packet.data.note.channel = channel;
			packet.data.note.note = p->data[0];
			packet.data.note.velocity = p->data[1];
			break;
	
		case 0xA0:
			packet.type = SND_SEQ_EVENT_KEYPRESS;
			packet.data.note.channel = channel;
			packet.data.note.note = p->data[0];
			packet.data.note.velocity = p->data[1];
			break;
		
		case 0xB0:
			packet.type = SND_SEQ_EVENT_CONTROLLER;
			packet.data.control.channel = channel;
			packet.data.control.param = p->data[0];
			packet.data.control.value = p->data[1];
			break;

		case 0xC0:
			packet.type = SND_SEQ_EVENT_PGMCHANGE;
			packet.data.control.channel = channel;
			packet.data.control.value = p->data[0];
			break;

		case 0xD0:
			packet.type = SND_SEQ_EVENT_CHANPRESS;
			packet.data.control.channel = channel;
			packet.data.control.value = p->data[0];
			break;
		
		case 0xE0:
			packet.type = SND_SEQ_EVENT_PITCHBEND;
			packet.data.control.channel = channel;
			packet.data.control.value = (int) p->data[0] | (int) p->data[1] << 7;
			packet.data.control.value -= 0x2000;
			break;

		default:
			msg(), this, "dropping ", p;
			break;
	}

	if (packet.type)
	{
		int r = snd_seq_event_output_direct (handle, &packet);
		if (r < 0)
			msg(), this, "sending ", p, " fails: ", strerror (-r);
	}
}

void
SeqOut::send_sysex (Packet * p)
{
	if (!p->payload)
	{
		msg(), this, ": ignoring SysEx without data: ", p;
		return;
	}
		
	int length = p->length;
	byte * data = (byte *) p->payload;

#if 0
	int written = 0;

	while (written < length)
	{
		int n = min (256, length - written);
		written += n;

		snd_seq_ev_set_sysex (&packet, n, data + written);

		int r = snd_seq_event_output_direct (handle, &packet);
		if (r < 0)
		{
			msg(), this, "sending event fails: ", strerror (-r);
			return;
		}
	}
#else
	byte f0 = 0xF0;
	snd_seq_ev_set_sysex (&packet, 1, &f0);
	int r = snd_seq_event_output_direct (handle, &packet);
	if (r < 0)
	{
		msg(), this, "sending F0 fails: ", strerror (-r);
		return;
	}

	snd_seq_ev_set_sysex (&packet, length, data);

	r = snd_seq_event_output_direct (handle, &packet);
	if (r < 0)
	{
		msg(), this, "sending sysex fails: ", strerror (-r);
		return;
	}
#endif
}

/* ////////////////////////////////////////////////////////// */

#include "ALSA_Seq.thunk.h"
#if PYTHUNK
	Class (ALSA::SeqClient, Plugin);
	Method (ALSA::SeqClient :: _subscribe, VA);
	Method (ALSA::SeqClient :: _unsubscribe, VA);
	Class (ALSA::SeqIn, ALSA::SeqClient);
	Class (ALSA::SeqOut, ALSA::SeqClient);
#endif

}; /* namespace ALSA */

PyObject * 
PyModule (char * name, PyObject * dict)
{
	PyObject * m = PyModule_New (name);
	PyDict_SetItemString (dict, name, m);
	return PyModule_GetDict (m);
}

void
init_ALSA (PyObject * dict)
{
	dict = PyModule ("ALSA", dict);

	new PyType<ALSA::SeqClient> ("SeqClient", dict);
	new PyType<ALSA::SeqIn> ("SeqIn", dict);
	new PyType<ALSA::SeqOut> ("SeqOut", dict);
}
