#ifndef _CYCLE_H_
#define _CYCLE_H_

#include "Tempo.h"
#include "FIFO.h"
#include "Recycler.h"

template <class T>
struct Interval
{
	T now, then, delta;

	Interval()
		{ now = then = delta = 0; }

	bool has (T t)
		{ return t >= then && t < now; }
};

class TX;

class _Cycle
{
	protected:
		Recycler recycler;

	public:
		friend class Flow;
		friend class MasterCycle;
		Flow * flow;

		Time::tick tick;
		Time::time time;

		_Cycle (Flow * f, int n)
			: recycler (n)
				{ flow = f; }

		/* ////////////////////////////////////// Packet management */
		
		void discard (Packet * p)
			{ recycler.push (p); }
		
		/* /////////////////////// Packet constructors */

		Packet * packet()
			{ return recycler.pop(); }

		Packet * packet (Event * e)
			{ 
				Packet * p = packet(); 
				(Event &) *p = *e;
				return p;
			}

		Packet * packet (Event::Type type, double t)
			{ 
				Packet * p = packet();
				p->type = type;
				p->t = t;
				return p;
			}

		Packet * packet (PyObject * args);

		/* //////////////////////////////////// Packet transmission */

		void push (Event & e, TX * tx)
			{ push (packet (&e), tx); }

		void push (Packet * p, TX * tx);
};

/* /////////////////////////////////////////////////////////////////// */

/* Cycle is managing regular interval-driven processing cycles, based on
 * the real time clock, the musical tempo or audio processing. */

class Cycle
: public _Cycle
{
	public:
		int id;
		enum {
			Time = 1,
			Tick = 2,
			Audio = 4,
			IO = 8,
			Script = 16
		};

		Interval<Time::tick> tick;
		Interval<Time::time> time;

		Tempo * tempo;
		Time::tick quarter;

		Time::tick time2tick (Time::time t)
			{ return tempo->map_time (t); }
		
		Time::time tick2time (Time::tick t)
			{ return tempo->map_tick (t); }
		
		Cycle (Flow * f, int _id, int n) 
			: _Cycle (f, n)
			 { id = _id; tempo = 0; quarter = 96; }

		/* /////////////////////////////////////////////////// Time steps */

		void step_time (Time::time dt)
			{
				time.then = time.now;
				time.now += dt;
				time.delta = dt;

				init_from_time();
			}

		void step_tick (Time::tick dt)
			{
				tick.then = tick.now;
				tick.now += dt;
				tick.delta = dt;

				init_from_tick();
			}

		void init_from_time()
			{
				tick.then = time2tick (time.then);
				tick.now = time2tick (time.now);
				tick.delta = tick.now - tick.then;
			}

		void init_from_tick()
			{
				time.then = tick2time (tick.then);
				time.now = tick2time (tick.now);
				time.delta = time.now - time.then;
			}
};

/* /////////////////////////////////////////////////////////////////// */

/* IOCycle manages event-driven processing cycles, like MIDI in and disk I/O. 
 */

class IOCycle
: public _Cycle
{
	public:
		IOCycle (Flow * f, int n)
			: _Cycle (f, n)
			{ }
};

#endif /* _CYCLE_H_ */
