#include "basics.h"
#include "Flow.h"
#include "Audio.h"
#include "Py.h"
#include "PyLock.h"

Flow::Flow()
: cycle (this, &tempo),
	run_script (&thread.pfds),
	rq_answer (&thread.pfds)
{
	run_script.set_nonblocking (Pipe::Read);
	pulse.last = now();
	audio_io = 0;
}

Flow::~Flow()
{
	if (audio_io)
		set_audio_io ((AudioIO *) 0);
	
	rt_events.quit();
	rtc.thread.quit();

	{
		PyUnlock _unlock;
		thread.quit();
	}

	for (int i = 0; i < plugins; ++i)
		plugins[i]->disconnect();

	decycle();

	msg(), this->as_py_object(), " is done.";
}

/* /////////////////////////////////////////////// AudioIO add, rm */ 

void 
Flow::set_audio_io (AudioIO * a)
{
	if (a && a->flow)
		throw RuntimeError() + a->as_py_object() + " is busy";

	if (audio_io)
		audio_io->quit();
	
	audio_io = 0;
	if (a == 0)
		return;

	a->set_flow (this);
	audio_io = a;
}

int
Flow::set_audio_io (PyObject * args)
{
	PyObject * _a;

	if (args && !PyArg_Parse (args, "O", &_a))
		throw blewit();

	if (_a == Py_None)
		_a = 0;
	else if (_a && !PyArg_Parse (args, "O!", Py<AudioIO>::type, &_a))
		return -1;

	AudioIO * a = &((Py<AudioIO> *) _a)->obj;

	set_audio_io (a);

	incref (_a);

	return 0;
}

PyObject * 
Flow::get_audio_io()
{
	return audio_io ? incref (audio_io->as_py_object()) : PyNone();
}

/* /////////////////////////////////////////////// Plugin add, rm */

void
Flow::add_plugin (Py<Plugin> * pp)
{
	plugins.append (pp);
	Plugin * p = &pp->obj;
	
	if (p->plugin.fd >= 0)
	{
		msg(), pp, " registered fd ", p->plugin.fd;
		rt_events.pfds.add_fd (p->plugin.fd, POLLIN);
	}
}

void
Flow::rm_plugin (Py<Plugin> * pp)
{
	plugins.remove (pp);
	Plugin * p = &pp->obj;

	if (p->plugin.fd >= 0)
		rt_events.pfds.remove_fd (p->plugin.fd);
}

/* /////////////////////////////////////////////// */

void
Flow::audio_cycle (int frames)
{
	// cycle.audio.step_audio (frames);

	for (int i = 0; i < plugins; ++i)
	{
		Plugin * p = plugins[i];
		
		if (p->plugin.flags & Cycle::Audio)
			p->cycle (cycle.audio);
	}

	Time::tick t = tempo.map_time (cycle.time.time.now);
	if (cycle.tick.tick.now - t < 2 * cycle.tick.quarter)
		thread.request.write_trigger ((void *) &cycle.tick);

	post_cycle();
}

/* /////////////////////////////////////////////// */

void
Flow::time_cycle()
{
	Lock _lock (mutex);

	process_requests();

	Time::time now = ::now();
	cycle.time.step_time (now - pulse.last);
	pulse.last = now;

	for (int i = 0; i < plugins; ++i)
	{
		Plugin * p = plugins[i];
		
		if (p->plugin.flags & Cycle::Time)
			p->cycle (cycle.time);
	}

	Time::tick t = tempo.map_time (cycle.time.time.now);
	if (cycle.tick.tick.now - t < 2 * cycle.tick.quarter)
		thread.request.write_trigger ((void *) &cycle.tick);

	post_cycle();
}

void
Flow::seek_to_time (Time::time t)
{
	cycle.time.time.now = t;
}

/* /////////////////////////////////////////////// */

void
Flow::tick_cycle()
{
	Lock _lock (mutex);

	cycle.tick.step_tick (cycle.tick.quarter);

	for (int i = 0; i < plugins; ++i)
	{
		Plugin * p = plugins[i];
		
		if (p->plugin.flags & Cycle::Tick)
		{
			PyLock __lock;
			p->cycle (cycle.tick);
		}
	}
	
	post_cycle();

	cycle.recycle();
}

/* /////////////////////////////////////////////// Script Plugin RX */

void
Flow::post_cycle()
{
	for (int i = 0; i < plugins; ++i)
	{
		Plugin * p = plugins[i];

		if (!(p->plugin.flags & Cycle::Script))
			continue;

		if (p->rx_packets() == 0)
			continue;
		
		run_script.write_trigger ((void *) p);
	}
}

void
Flow::script_cycle()
{
	Plugin * pp [32];
	int n;
	
	while ((n = run_script.read (&pp, 32 * sizeof (Plugin *))) > 0)
	{	
		n /= sizeof (Plugin *);

		for (int j = 0; j < n; ++j)
		{
			Lock _lock (mutex);

			Plugin * p = pp[j];
			for (int i = 0; i < plugins; ++i)
			{
				if (plugins[i] != p)
					continue;

				PyLock _lock;
				p->event (cycle.tick);
				break;
			}
		}
	}
}

/* //////////////////////////////////////////////////// Packet cache mgmt */

/* fetches all trashed packets and queues them up, then refills the 
 * cycle caches 
 */
void
MasterCycle::recycle()
{
	recycle_from (tick.recycler);
	recycle_from (time.recycler);
	recycle_from (tick_io.recycler);
	recycle_from (time_io.recycler);
	recycle_from (audio.recycler);

	recycle_to (tick.recycler);
	recycle_to (time.recycler);
	recycle_to (tick_io.recycler);
	recycle_to (time_io.recycler);
	recycle_to (audio.recycler);
}

void 
MasterCycle::recycle_from (Recycler & r)
{
	while (r.trash.have())
	{
		Packet * p = r.trash.pop();
		p->free();
		recycled.push (p);
	}
}

void 
MasterCycle::recycle_to (Recycler & r)
{
	while (r.cache.free() && recycled.have())
		r.cache.push (recycled.pop());
}

void
Flow::decycle()
{
	for (int i = 0; i < plugins; ++i)
	{
		Plugin * p = plugins[i];
		
		for (int j = 0; j < p->plugin._rx->obj; ++j)
		{
			RX * rx = p->rx(j);

			while (rx->fifo.have())
				rx->fifo.pop()->free();
		}
	}

	/* no more packets within the flow, so all packets' ownership is back 
	 * at recycler */
	cycle.tick.recycler.refill();
	cycle.time.recycler.refill();
}

/* ////////////////////////////////////////////// */

void
Flow::handle_io (IOCycle & c, int fd)
{
	Lock _lock (mutex);

	Time::time dt = ::now() - pulse.last;
	
	c.time = cycle.time.time.now + dt;
	c.tick = tempo.map_time (c.time);

	for (int i = 0; i < plugins; ++i)
	{
		Plugin * p = plugins[i];
		
		if (p->plugin.fd == fd)
			p->handle_io (fd, c);
	}
}

void
Flow::handle_thread_event (Thread * t, int fd)
{
	if (fd < 0) /* timed out */
		return;

	if (fd == run_script.fd[Pipe::Read])
	{
		if (t != &thread)
			throw AssertionError() + "run_script not in script thread";
		script_cycle();
		return;
	}

	if (t == &rt_events)
		handle_io (cycle.time_io, fd);
	else if (t == &thread)
		handle_io (cycle.tick_io, fd);
	else
		throw AssertionError() + "thread is not supposed to signal events";
}

void
Flow::handle_trigger (Thread * t, void * trigger)
{
	if (trigger == (void *) &cycle.tick)
		tick_cycle();
	else
		msg(), "trigger: ", trigger;
}

/* //////////////////////////////////////////// */

/* outside, passing the request from Python to the graph */

void
Flow::request (Request * r)
{
	{
		PyUnlock _unlock;
		requests.push (r);
		rtc.thread.request.write_trigger ((void *) r);
		rq_answer.read_trigger();
	}

	if (r->blown.excuse)
		throw r->blown;
}

/* inside, processing the request */

void
Flow::process_requests()
{
	while (requests.have())
		handle_request (requests.pop());
}

void 
Flow::handle_request (Request * r)
{
	try 
	{
		r->exec (this);
	}
	catch (blewit & b)
	{
		r->blown = b;
	}

	rq_answer.write_trigger (r);
}


