#include "basics.h"
#include "Py.h"

#include "Jack.h"

static int 
_jack_buffer_size (jack_nframes_t n, void * self)
{
	return ((Jack *) self)->jack_buffer_size (n);
}

static int 
_jack_sample_rate (jack_nframes_t n, void * self)
{
	return ((Jack *) self)->jack_sample_rate (n);
}

static void 
_jack_shutdown (void * self)
{
	((Jack *) self)->jack_shutdown();
}

static int 
_jack_process (jack_nframes_t n, void * self)
{
	return ((Jack *) self)->jack_process (n);
}

#if 0
static void 
_jack_verror (const char * fmt, ...)
{
	char * s = 0;
	va_list args;
	
	va_start (args, fmt);
	vasprintf (&s, fmt, args);
	va_end (args);
	
	msg(), "Jack says: ", s;

	free (s);
}
#endif

static void 
_jack_error (const char * error)
{
	msg(), "Jack says: ", error;
}

/* //////////////////////////////////////////////////////////////////////// */

Jack::Jack()
{
	client = 0;

	jackd_is_down = true;
	first_run = true;

	jack_set_error_function (_jack_error);
}

Jack::~Jack()
{
	if (client)
		jack_client_close (client);
}

void
Jack::jack_shutdown()
{
	msg(), "jackd went down.";
	jackd_is_down = true;
}

int
Jack::jack_process (int frames)
{
	if (first_run)
	{
		first_run = false;
		thread.set_specifics();
	}
	
	if (flow == 0)
		return 0;

	try
	{
		cycle_tx (frames);
		flow->audio_cycle (frames);
		cycle_rx (frames);
	}
	catch (blewit & b)
	{
		msg(), "audio cycle blew it: ", b.what();
	}

	return 0;
}

void
Jack::cycle_tx (int frames)
{
	Cycle & c = flow->cycle.audio;

	for (int i = 0; i < plugin._rx->obj.length; ++i)
	{
		Packet * p = c.packet();
		sample * s = ((JackIn *) tx(i))->get_samples (frames);
		p->set_audio (s, frames);
		c.push (p, tx(i));
	}
}

void
Jack::cycle_rx (int frames)
{
	Cycle & c = flow->cycle.audio;

	for (int i = 0; i < plugin._rx->obj.length; ++i)
	{
		Packet * p = rx(i)->get_audio (&c);
		if (p == 0)
			continue;
		
		sample * s = ((JackOut *) rx(i))->get_samples (frames);
		memcpy (s, p->payload, frames * sizeof (sample));
		c.discard (p);
	}
}

void
Jack::update_flow()
{
	if (flow == 0)
		return;
}

void
Jack::set_flow (Flow * f)
{
	flow = f;
}

/* ////////////////////////////////////////////////// */

void
Jack::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"name", "i", "o", 0};
	char * name;
	int i, o;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "sii", keys, &name, &i, &o))
		throw blewit();
	
	client = jack_client_new (name);
	
	if (client == 0)
		throw IOError() + " can't register with the Jack server. Is it up?";

	jack_on_shutdown (client, _jack_shutdown, this);

	sample_rate = jack_get_sample_rate (client);
	buffer_size = jack_get_buffer_size (client);
	
	jack_set_buffer_size_callback (client, _jack_buffer_size, this);
	jack_set_sample_rate_callback (client, _jack_sample_rate, this);

	jack_set_process_callback (client, _jack_process, this);

	for (int j = 0; j < i; ++j)
	{
		char s[64];
		sprintf (s, "in_%d", j);
		new_tx<JackIn>(s)->setup (client, s);
	}

	for (int j = 0; j < o; ++j)
	{
		char s[64];
		sprintf (s, "out_%d", j);
		new_rx<JackOut>(s)->setup (client, s);
	}

	if (jack_activate (client))
		throw IOError() + "jackd refuses to activate this client.";

	jackd_is_down = false;
}

PyObject * 
Jack::_connect (PyObject * args)
{
	char * s, * d;
	if (!PyArg_ParseTuple (args, "ss", &s, &d))
		return 0;

	return PyInt (jack_connect (client, s, d));
}

#include "Jack.thunk.h"

#if PYTHUNK
	Class (AudioIO, Plugin)
	Class (JackIn, TX)
	Class (JackOut, RX)
	Class (Jack, AudioIO)
	Method (Jack :: _connect, VA)
#endif

void
init_Jack (PyObject * dict)
{
	new PyType<AudioIO> ("AudioIO", dict);
	new PyType<JackIn> ("JackIn", dict);
	new PyType<JackOut> ("JackOut", dict);
	new PyType<Jack> ("Jack", dict);
}
