#include "basics.h"
#include "MIDIOut.h"
#include "MIDIIn.h"
#include "Py.h"

MIDIIn::MIDIIn()
{
	f8 = 1;
	fe = 0; /* ignore active sense */

	plugin.fd = -1;
	event.type = Event::Invalid;
	running = 0;
}

MIDIIn::~MIDIIn()
{
	if (plugin.fd >= 0) close (plugin.fd);
}

void
MIDIIn::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", 0};
	char * path = "/dev/snd/midiC3D0";
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|s", keys, &path))
		throw blewit();

	plugin.fd = open (path, O_RDONLY);
	if (plugin.fd < 0)
		throw OSError() + path;

	if (fcntl (plugin.fd, F_SETFL, O_NONBLOCK) == -1)
		throw OSError() + "setting nonblocking mode on file";

	new_tx<TX> ("midi");
}

/* //////////////////////////////////////////////// */

void
MIDIIn::start_event (byte status)
{
	data.length = 0;
	event.type = (Event::Type) status;

	if (status < 0x80) /* not locked on. */
		bytes = 0;
	else if (status < 0xC0)	/* note off/on, note touch & control change */
		bytes = 2,
		running = status;
	else if (status <= 0xDF) /* program change & channel touch */
		bytes = 1,
		running = status;
	else if (status < 0xF0) /* pitch wheel */
		bytes = 2,
		running = status;
	else if (status == 0xF0)
		bytes = -1;
	else if (status == 0xF1) /* SMPTE time code */
		bytes = 1;
	else if (status == 0xF2) /* song position */
		bytes = 2;
	else if (status == 0xF3) /* song select */
		bytes = 1;
	else
		bytes = 0;
	
	event.t = cycle->tick;

	if (bytes)
		expect = Data;
	else
		dispatch(); /* status byte only event */
}

void
MIDIIn::dispatch()
{
	if (event.type == Event::Invalid)
		throw AssertionError();
	if (event.type == 0xF0)
	{
		Packet * p = cycle->packet (&event);
		p->set_data (data.items, data.length);
		cycle->push (p, tx(0));
	}
	else
	{
		event.data[0] = data[0];
		event.data[1] = data[1];

		cycle->push (event, tx(0));
	}

	event.type = Event::Invalid;
	expect = Status;
}

void
MIDIIn::parse (byte b)
{
	if (b >= 0xF8) /* one-byte realtime message */
	{
		if (b == 0xF8 && !f8)
			return;
		if (b == 0xFE && !fe)
			return;
		Packet * p = cycle->packet ((Event::Type) b, cycle->tick);
		cycle->push (p, tx(0));
		return;
	}

	if (b & 0x80) /* status byte */
	{
		/* event terminated by new status. */
		if (expect == Data) 
		{
			/* append the EOX byte. */
			if (event.type == 0xF0 && b == 0xF7)
			{
				data.append (b);
				dispatch();
				return;
			}

			dispatch();
		}

		/* new status, new event. */
		start_event (b);
		return;
	}

	if (expect == Status) 
	{
		if (!running) 
			return;

		start_event (running);
	}

	if (expect == Data)
	{
		data.append (b);

		if (--bytes == 0)
			dispatch();
	}
}

void
MIDIIn::handle_io (int fd, IOCycle & c)
{
	cycle = &c;

	byte buffer[256];

	while (1)
	{
		int n = read (fd, buffer, 256 * sizeof (byte));
		if (n <= 0)
			break;

		for (int i = 0; i < n; ++i)
			parse (buffer[i]);
	}
}

/* /////////////////////////////////////////// */

MIDIOut::MIDIOut()
{
	plugin.flags = Cycle::Time;
	fd = -1;
	bandwidth = 31250. / 10.;
}

MIDIOut::~MIDIOut()
{
	if (fd >= 0) close (fd);
}

void
MIDIOut::reset()
{
	current.packet = 0;
	current.index = -1;
	running = 0;
	writable = 0;
}

bool
MIDIOut::prepare (Packet * p)
{
	/* MIDI channel message payload length by status nybble */
	static int len [] = {0,0,0,0, 0,0,0,0, 2,2,2,2, 1,1,2, 0};
	
	if (p->type >= 8 && p->type <= 0xE)
		current.status = (p->type << 4) | current.channel;
	else if (p->type >= 0x80 && p->type <= 0xEF)
		current.status = p->type;
	else if (p->type == 0xF0)
		current.status = p->type;
	else
		return false;
	
	if (p->payload)
	{
		current.length = p->length;
		current.data = (byte *) p->payload;
	}
	else
	{
		current.length = len[current.status >> 4];
		current.data = p->data;
	}

	return true;
}

int
MIDIOut::enqueue (byte * buffer, int nbytes, Cycle & c)
{
	Packet * p = current.packet;
	int written = 0;

	while (written < nbytes)
	{
		/* done with the current packet? */
		if (p && current.index >= current.length)
		{
			c.discard (p);
			current.packet = p = 0;
		}

		/* fetch next packet */
		if (p == 0)
		{
			current.index = -1;
			p = current.packet = next_packet (c.tick.now, &current.channel);
					
			if (p == 0)
				break;

			if (!prepare (p))
			{
				c.discard (p);
				p = 0;
				continue;
			}
		}

		if (current.index == -1)
		{
			if (current.status >= 0xF0)
				buffer[written++] = current.status;
			else if (current.status != running)
				buffer[written++] = running = current.status;
		}
		else
		{
			buffer[written++] = current.data[current.index];
		}

		++current.index;
	}

	return written;
}

void
MIDIOut::cycle (Cycle & c)
{
	writable += c.time.delta * bandwidth;
	int nbytes = (int) writable;
	writable -= nbytes;

	byte b [nbytes];
	int written = 0;

	Time::tick dt = c.quarter / 24.;
	Time::tick t = c.tick.then - fmod (c.tick.then, dt) + dt;

	while (t < c.tick.now && written < nbytes)
	{
		b[written++] = 0xF8;
		t += dt;
	}

	while (written < nbytes)
	{
		int w = enqueue (b + written, nbytes, c);
		if (w == 0)
			break;
		written += w;
	}

	if (1 && written > 1)
	{
		message m;
		m, "MIDI out: ";
		m.bytes (b, written);
	}

	write (fd, b, written);
}

/* ////////////////////////////////////////////////////////// */

void
MIDIOut::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", 0};
	char * path = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &path))
		throw blewit();

	fd = open (path, O_WRONLY);
	if (fd < 0)
		throw OSError() + path;

	for (int i = 0; i < 16; ++i)
	{
		char s[8];
		sprintf (s, "%d", i + 1);
		new_rx<RX> (s);
	}
}

#include "MIDI.thunk.h"
#if PYTHUNK
	Class (MIDIOut, Plugin)
	Class (MIDIIn, Plugin)
#endif

void 
init_MIDI (PyObject * dict)
{
	new PyType<MIDIOut> ("MIDIOut", dict);
	new PyType<MIDIIn> ("MIDIIn", dict);
}
