#ifndef _POLL_H_
#define _POLL_H_

#include <sys/poll.h>
#include "List.h"

class PFDs;

class Pipe
{
	public:
		int fd[2];

		enum {
			Read = 0,
			Write = 1
		};

		inline Pipe (PFDs * p);

		void init()
			{
				if (pipe (fd))
					throw OSError(errno) + "creating Pipe";
				
				set_nonblocking (Write);
			}

		~Pipe()
			{
				close (fd[0]);
				close (fd[1]);
			}

		void set_nonblocking (int i)
			{
				if (fcntl (fd[i], F_SETFL, O_NONBLOCK) == -1)
					throw OSError() + "setting nonblocking mode on Pipe";
			}

		inline int write (void * x, int n)
			{
				int w = ::write (fd[Write], x, n);
				
				if (w <= 0)
					throw OSError() + "writing to Pipe"; 
				
				return w;
			}

		inline int read (void * x, int n)
			{
				return ::read (fd[Read], x, n);
			}

		inline void write_trigger (void * trigger)
			{ write (&trigger, sizeof (void *)); }

		inline void * read_trigger()
			{
				void * trigger;
				read (&trigger, sizeof (void *));
				return trigger;
			}
};

/* ////////////////////////////////////////////// */

class PFDs
{
	public:
		List<struct pollfd, 16> pfds;

		PFDs() { }

		/* returns the file descriptor that triggered or -1 */
		int poll (int ms)
			{
				/* FIXME: this isn't safe against adding to the pfds array while we
				 * are in ::poll() -- the pfds array can be realloc()ated to another
				 * spot in memory. the short-term fix here is to grab the number of
				 * currently monitored fds before poll and hope that the granularity
				 * of pfds[] isn't exceeded. */
				int n = (int) pfds;
				
				for (int i = 0; i < n; ++i)
					pfds[i].revents = 0;

				int r = ::poll (pfds, n, ms);

				if (r < 0)
				{
					if (errno == EINTR)
						message(), "EINTR in poll()";
					else
						throw IOError (errno) + "poll()";
					errno = 0;
				}

				for (int i = 0; i < n; ++i)
				{
					if (pfds[i].revents & ~(POLLIN | POLLOUT))
						msg(), pfds[i].fd, " revents=", (void *) (long) pfds[i].revents;

					if (pfds[i].revents & POLLERR)
						throw IOError() + " poll error.";
					if (pfds[i].revents & POLLHUP)
						throw IOError() + " poll: fd " + pfds[i].fd + " hung up.";
					if (pfds[i].revents & (POLLIN | POLLOUT))
						return pfds[i].fd;
				}

				return -1;
			}

		void add_pipe (Pipe * pipe)
			{
				add_fd (pipe->fd[Pipe::Read], POLLIN);
			}

		void add_fd (int fd, int events)
			{
				int i = pfds.append();
				pfds[i].fd = fd;
				pfds[i].events = events;
			}

		void remove_fd (int fd)
			{
				int i;

				for (i = 0; i < pfds; ++i)
				{
					if (pfds[i].fd == fd)
						break;
				}

				if (i == pfds)
					throw LookupError() + "fd not found: " + fd;
				
				pfds.remove_items (i, 1);
			}

		operator int ()
			{ return pfds; }
};

inline
Pipe::Pipe (PFDs * pfds)
{
	init();
	pfds->add_pipe (this);
}

#endif /* _POLL_H_ */
