#ifndef _THREAD_H_
#define _THREAD_H_

#include "Poll.h"

#define PTHR(call) \
	{ \
		int e = call; \
		if (e) throw OSError (e) + ": " + #call; \
	}

class Thread
{
	public:
		static pthread_key_t key;
		pthread_t id;

		char * name;
		int interval; /* ms for poll() */

		PFDs pfds;
		Pipe request;

		/* support for PyLock */
		PyThreadState * tstate;
		PyInterpreterState * interp;

		Thread()
			: request (&pfds)
			{ 
				name = "<unknown>";
				id = 0;
				interval = 1000;
				tstate = 0;
				interp = PyThreadState_Get()->interp;
				tstate = PyThreadState_New (interp);
			}

		~Thread() 
			{ 
				id = 0; 

				if (tstate)
				{
					PyThreadState_Clear (tstate);
					PyThreadState_Delete (tstate);
				}
			}

		void quit()
			{
				if (id == 0) 
					return;
				
				request.write_trigger (0);

				if (id == 0) 
					return;
				
				void * r;
				PTHR (pthread_join (id, &r));
			}

		inline static Thread * current()
			{
				return (Thread *) pthread_getspecific (key);
			}

	public:
		void set_specifics()
			{
				PTHR (pthread_setspecific (key, (void *) this));
			}
};
	
template <class T>
class ThreadT
: public Thread
{
	public:
		T * obj;

		/* poll for events, calling handle_trigger if a trigger has been 
		 * passed in, handle_thread_event else. */
		void thread()
			{
				while (1)
				{
					int fd = pfds.poll (interval);

					if (0) {
						message m;
						m, "poll: ";
						for (int i = 0; i < pfds; ++i)
							m, pfds.pfds[i].fd, " ";
						m, "= ", fd;
					}
					
					if (fd != request.fd[Pipe::Read])
						obj->handle_thread_event (this, fd);
					else
					{
						void * t = request.read_trigger();
						if (t == 0)
							return;
						obj->handle_trigger (this, t);
					}
				}
			}

		/* //////////////////////////////////// */

		static void * _thunk (void * self)
			{
				ThreadT<T> * t = (ThreadT<T> *) self;
				
				((Thread *) t)->set_specifics();

				/* for debugging, we sometimes want SIGABRT raised instead of 
				 * gracious handling, that's when the if (1) is quick to change */
				if (1)
				{
					try 
					{
						t->thread();
						message(), "thread finished.";
					} 
					catch (blewit & b) 
					{
						message(), "blew it: ", b.s;
					}
				}
				else
				{
					t->thread();
					message(), "thread finished.";
				}
				
				t->id = 0;
				return 0;
			}

		/* ///////////////////////////////////////////////////////////////////// */
		
		void start (T * t, char * _name, float priority = 0, bool realtime = false)
			{
				obj = t;
				name = _name;

				struct sched_param sched;

				while (1)
				{
					int policy = realtime ? SCHED_FIFO : SCHED_OTHER;
					
					float p0 = sched_get_priority_min (policy);
					float p1 = sched_get_priority_max (policy);
					
					sched.sched_priority = (int) (p0 + priority * (p1 - p0));
				
					pthread_attr_t attr;
					
					PTHR (pthread_attr_init (&attr));
					PTHR (pthread_attr_setschedpolicy (&attr, policy));
					PTHR (pthread_attr_setschedparam (&attr, &sched));
					PTHR (pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM));
					
					int e = pthread_create (&id, &attr, _thunk, this);

					if (e == 0)
						break;

					msg(), "cannot start thread ", realtime ? "RT " : "",
							"prio ", sched.sched_priority, ": ", strerror (e);

					/* if realtime or priority is a problem, resort to normal */
					if (realtime)
						realtime = false;
					else if (priority != 0)
						priority = 0;
					else
						throw OSError (e) + " cannot start thread.";
				}

				msg(), "thread ", name, " started as ", (int) id,
						", prio=", realtime ? "FIFO:" : "OTHER:", sched.sched_priority;
			}
};

#undef PTHR

#endif /* _THREAD_H_ */
