#! /usr/bin/env python
# --------------------------------------------------------------------------------
# generates thunk code for calling C++ class members from python, with
# locking and exception handling around the C++ call

import sys, re

class Cls:
	def __init__ (self, base):
		self.doc = "doc"
		self.base = base
		self.methods = []
		self.getsets = {}
		self.sequence = {}

clslist = []
classes = {}

def thunk_class (args):
	c, base = args
	clslist.append (c)
	classes[c] = Cls (base)

# -------------------------------------------------------------- SEQUENCE
# sequence methods are always wrapped in a scoped Lock (self->mutex)
SQ_LENGTH = """
static int
%s_sq_length (Py<%s> * self)
{
	Lock _lock (self->obj.mutex);
	return self->obj.sq_length();
}"""

SQ_ITEM = """
static PyObject *
%s_sq_item (Py<%s> * self, int i)
{
	Lock _lock (self->obj.mutex);
	try {
		return self->obj.sq_item (i);
	} catch (blewit & b) {
		b.raise();
	}
	return 0;
}"""

class Sequence:
	struct = (
		("inquiry", "length"),
		("binaryfunc", "concat"),
		("intargfunc", "repeat"),
		("intargfunc", "item"),
		("intintargfunc", "slice"),
		("intobjargproc", "ass_item"),
		("intintobjargproc", "ass_slice"),
		("objobjproc", "contains"),
		("binaryfunc", "inplace_concat"),
		("intargfunc", "inplace_repeat"),
	)
	def length (c):
		print SQ_LENGTH % (c, c)
	def item (c):
		print SQ_ITEM % (c, c)

def thunk_sq (args):
	sq, c = args
	classes[c].sequence[sq] = "%s_sq_%s" % (c, sq)
	f = Sequence.__dict__[sq]
	f (c)
	
# ------------------------------------------------------------------ METHOD

METH_VA = """
static PyObject * 
%s_%s (Py<%s> * self, PyObject * args)
{
	try {
		return self->obj.%s (args);
	} catch (blewit & b) {
		b.raise();
	}
	return 0;
}""" 

METH_KW = """
static PyObject * 
%s_%s (Py<%s> * self, PyObject * args, PyObject * kwds)
{
	try {
		return self->obj.%s (args, kwds);
	} catch (blewit & b) {
		b.raise();
	}
	return 0;
}"""

def thunk_method (args):
	c, m, a = args
	name = m
	m = m[0] == '_' and m[1:] or m
	classes[c].methods.append ((m, a))
	if a == 'VA': 
		print METH_VA % (c, m, c, name)
	elif a == 'KW': 
		print METH_KW % (c, m, c, name)

# --------------------------------------------------------------- GETSET

GET = """
static PyObject * 
%s_get_%s (Py<%s> * self)
{
	try {
		return self->obj.get_%s();
	} catch (blewit & b) {
		b.raise();
	}
	return 0;
}"""

SET = """
static int
%s_set_%s (Py<%s> * self, PyObject * args)
{
	try {
		return self->obj.set_%s (args);
	} catch (blewit & b) {
		b.raise();
	}
	return -1;
}"""

def thunk_gset (args):
	gs, c, p = args
	classes[c].getsets[p] = gs
	if "G" in gs:
		print GET % (c, p, c, p)
	if "S" in gs:
		print SET % (c, p, c, p)

def thunk_getset (args):
	c, p = args
	thunk_gset (("GS", c, p))

GETSET = """
	{
		"%s", (getter) %s, (setter) %s,
		"%s"
	},"""

def dump_getsets (c):
	getsets = classes[c].getsets
	print """
struct PyGetSetDef %s_getsets [] = {""" % c,
	for p in getsets:
		gs = getsets[p]
		d = "doc"
		g = "G" in gs and ("%s_get_%s" % (c, p)) or "0"
		s = "S" in gs and ("%s_set_%s" % (c, p)) or "0"
		print GETSET % (p, g, s, d),
	print """
	{ }
};"""

# -------------------------------------------------------------------------

NAME = r'\s*(\w*:*\w*)\s*'
STR = r'\s*"(.*?)"\s*'
SQ = r'\s*sq_(\w*)\s*'
ARGS = r'\s*(\w*)\s*'

thunks = (
	(thunk_class, re.compile (r'\s*Class\s*\(' + NAME + ',' + NAME + r'\)')),
	(thunk_method, re.compile (r'\s*Method\s*\(' + NAME + ' :: ' + NAME + ',' + ARGS + r'\)')),
	(thunk_gset, re.compile (r'\s*([GS])et\s*\(' + NAME + ' :: ' + NAME + r'\)')),
	(thunk_getset, re.compile (r'\s*GetSet\s*\(' + NAME + ' :: ' + NAME + r'\)')),
	(thunk_sq, re.compile (SQ + '\(' + NAME + '\)')),
)

# -------------------------------------------------------------------------

print "/* Generated for '%s' by pythunk.py */" % sys.argv[1]

lines = open (sys.argv[1]).readlines()

try:
	i = lines.__iter__()
	while 1:
		l = i.next()
		if not l.startswith ("#if PYTHUNK"): continue
		while 1:
			l = i.next()
			if l.startswith ("#endif"): break
			for f, exp in thunks:
				m = exp.match (l)
				if not m: continue
				break
			if m:
				f (m.groups())
			elif l.strip():
				print >> sys.stderr, "Can't parse:", l
				sys.exit (1)

except StopIteration: pass

METH_CALL = {
	'VA': "METH_VARARGS",
	'KW': "METH_VARARGS | METH_KEYWORDS",
}

for c in clslist:
	print """
PyTypeObject * Py<%s>::type = 0;""" % (c,)

	k = classes[c]
	if k.methods:
		print """
static PyMethodDef %s_methods[] = {""" % c,
		for m, a in k.methods:
			d = "doc"
			a = METH_CALL[a]
			print """
	{"%s", (PyCFunction) %s_%s, %s, "%s"},""" % (m, c, m, a, d),
		print """
	{ }
};"""
	if k.sequence:
		print """
static PySequenceMethods
%s_as_sequence = {""" % c
		s = []
		for cast, name in Sequence.struct:
			call = k.sequence.get (name, "0")
			s.append ("(%s) %s" % (cast, call))
		print "\t" + ",\n\t".join (s),
		print """
};"""
				
	if k.getsets:
		dump_getsets (c)

	print """
void
PyType<%s>::__init__ (PyObject * dict)
{""" % c,
	
	if k.base != "None":
		print """
	tp_base = Py<%s>::type;""" % k.base,

	print """
	tp_doc = "%s";""" % k.doc,

	if k.methods: print """
	tp_methods = %s_methods;""" % (c),
	
	if k.sequence: print """
	tp_as_sequence = &%s_as_sequence;""" % (c),
	
	if k.getsets: print """
	tp_getset = %s_getsets;""" % (c),

	print """
}"""
